-- ada.adb

with Ada.Real_Time;               use Ada.Real_Time;
with Registers;                   use Registers;

package body LCD is

   DISP_INIT   : Constant Byte := 16#30#;
   DISP_4BITS  : Constant Byte := 16#20#;
   DISP_ON     : Constant Byte := 16#0C#;
   DISP_OFF    : Constant Byte := 16#08#;
   DISP_CLR    : Constant Byte := 16#01#;
   DISP_EMS    : Constant Byte := 16#06#;
   DISP_CONFIG : Constant Byte := 16#28#;

   LCD_RS   : constant Word := 2**2;
   LCD_RW   : constant Word := 2**8;
   LCD_E    : constant Word := 2**9;
   LCD_BKLT : constant Word := 2**15;

   DD_RAM_ADDR  : constant Byte := 16#00#;
   DD_RAM_ADDR2 : constant Byte := 16#40#;
   DD_RAM_ADDR3 : constant Byte := 16#14#;
   DD_RAM_ADDR4 : constant Byte := 16#54#;

   LCD_BSRR_XOR : constant Word := 16#00F0_0000#; -- inverts LO-going D4..D7


   procedure Wait_Until(S : Time_Span) is
      T : Constant Time := Clock + S;
   begin
      while Clock < T loop
         null;
      end loop;
   end Wait_Until;

   procedure LCD_Hi (Mask : Word) with Inline is
   begin
      GPIOE.BSRR := Mask;
   end LCD_Hi;

   procedure LCD_Lo (Mask : Word) with Inline is
   begin
      GPIOE.BSRR := Shift_Left (Mask, 16);
   end LCD_Lo;

   procedure LCD_strobe is
   begin
      LCD_Hi(LCD_E);
      Wait_Until(Nanoseconds(500));
      LCD_Lo(LCD_E);
   end LCD_strobe;

   procedure LCD_Write_4x2BitsX(C : Character) is
      gpio32 : Bits_32 with Size => 32;
      c8     : Bits_8 with Size => 8;
   begin
      Wait_Until(Microseconds(50));
      c8 := B8_From_B(Character'Pos(C));
      gpio32 := B32_From_W(GPIOE.ODR);
      gpio32(4..7) := c8(4..7);
      GPIOE.ODR := W_From_B32(gpio32);

      LCD_Strobe;

      gpio32(4..7) := c8(0..3);
      GPIOE.ODR := W_From_B32(gpio32);

      LCD_Strobe;

   end LCD_Write_4x2BitsX;

   procedure LCD_Write_4x2Bits(C : Character) is
      gpio32 : Bits_32 with Size => 32;
      gpiow  : Word;
      c8     : Bits_8 with Size => 8;
   begin
      Wait_Until(Microseconds(50));
      c8 := B8_From_B(Character'Pos(C));
      gpio32(0..31) := (others => False);
      gpio32(4..7) := c8(4..7);    -- Hi-going
      gpio32(20..23) := c8(4..7);  -- Lo-going
      gpiow := W_From_B32(gpio32);
      GPIOE.BSRR := gpiow xor LCD_BSRR_XOR;

      LCD_Strobe;

      gpio32(4..7) := c8(0..3);    -- Hi-going
      gpio32(20..23) := c8(0..3);  -- Lo-going
      gpiow := W_From_B32(gpio32);
      GPIOE.BSRR := gpiow xor LCD_BSRR_XOR; -- invert any Lo-going to 1s

      LCD_Strobe;

   end LCD_Write_4x2Bits;

   procedure LCD_Write_4x2BitsX(D : Byte) is
      gpio32 : Bits_32 with Size => 32;
      b8     : Bits_8 with Size => 8;
   begin
      Wait_Until(Microseconds(50));
      b8 := B8_From_B(D);
      gpio32 := B32_From_W(GPIOE.ODR);
      gpio32(4..7) := b8(4..7);
      GPIOE.ODR := W_From_B32(gpio32);

      LCD_Strobe;

      gpio32(4..7) := b8(0..3);
      GPIOE.ODR := W_From_B32(gpio32);

      LCD_Strobe;

   end LCD_Write_4x2BitsX;

   procedure LCD_Write_4x2Bits(D : Byte) is
      gpio32 : Bits_32 with Size => 32;
      gpiow  : Word;
      b8     : Bits_8 with Size => 8;
   begin
      Wait_Until(Microseconds(50));
      b8 := B8_From_B(D);
      gpio32(0..31) := (others => False);
      gpio32(4..7) := b8(4..7);    -- Hi-going
      gpio32(20..23) := b8(4..7);  -- Lo-going
      gpiow := W_From_B32(gpio32);
      GPIOE.BSRR := gpiow xor LCD_BSRR_XOR;

      LCD_Strobe;

      gpio32(4..7) := b8(0..3);    -- Hi-going
      gpio32(20..23) := b8(0..3);  -- Lo-going
      gpiow := W_From_B32(gpio32);
      GPIOE.BSRR := gpiow xor LCD_BSRR_XOR; -- invert any Lo-going to 1s

      LCD_Strobe;

   end LCD_Write_4x2Bits;

   procedure LCD_Write_4x1BitsX(D : Byte) is
      gpio32 : Bits_32 with Size => 32;
      b8     : Bits_8 with Size => 8;
   begin
      Wait_Until(Microseconds(50));
      b8 := B8_From_B(D);
      gpio32 := B32_From_W(GPIOE.ODR);
      gpio32(4..7) := b8(4..7);
      GPIOE.ODR := W_From_B32(gpio32);

      LCD_Strobe;
   end LCD_Write_4x1BitsX;

   procedure LCD_Cmd(Cmd : Byte) is
   begin
      LCD_Lo(LCD_RS);
      LCD_Write_4x2Bits(Cmd);
   end LCD_Cmd;

   procedure LCD_Init is

      procedure LCD_Write_4x1Bits(D : Byte) is
         gpio32 : Bits_32 with Size => 32;
         gpiow  : Word;
         b8     : Bits_8 with Size => 8;
      begin
         b8 := B8_From_B(D);
         gpio32(0..31) := (others => False);
         gpio32(4..7) := b8(4..7);    -- Hi-going
         gpio32(20..23) := b8(4..7);  -- Lo-going
         gpiow := W_From_B32(gpio32);
         GPIOE.BSRR := gpiow xor LCD_BSRR_XOR; -- invert any Lo-going to 1s

         LCD_Strobe;
      end LCD_Write_4x1Bits;

   begin
      LCD_Lo(LCD_E + LCD_RW + LCD_RS);
--      LCD_Lo(LCD_RW);
--      LCD_Lo(LCD_RS);

      LCD_Hi(LCD_BKLT);

      Wait_Until(Milliseconds(20));
      LCD_Write_4x1Bits(DISP_INIT);
      Wait_Until(Milliseconds(5));
      LCD_Write_4x1Bits(DISP_INIT);
      Wait_Until(Milliseconds(1));
      LCD_Write_4x1Bits(DISP_INIT);
      Wait_Until(Milliseconds(1));
      LCD_Write_4x1Bits(DISP_4BITS);
      Wait_Until(Milliseconds(1));

      LCD_Cmd(DISP_CONFIG);
      LCD_Cmd(DISP_OFF);
      LCD_Cmd(DISP_CLR);
      Wait_Until(Milliseconds(2));  -- undocumented but required!
      LCD_Cmd(DISP_EMS);
      LCD_Cmd(DISP_ON);
   end LCD_Init;

   procedure LCD_Clear is
   begin
      LCD_Cmd(DISP_CLR);
      Wait_Until(Milliseconds(2));  -- undocumented but required!
   end LCD_Clear;

   procedure LCD_Goto(X : Byte; Y : Byte) is
      Base : Byte;
   begin
      case Y is
         when 0 => Base := DD_RAM_ADDR;
         when 1 => Base := DD_RAM_ADDR2;
         when 2 => Base := DD_RAM_ADDR3;
         when 3 => Base := DD_RAM_ADDR4;
         when others => Base := DD_RAM_ADDR;
      end case;

      LCD_Cmd(Base + X + 16#80#);  -- addr command
   end LCD_Goto;

   procedure LCD_Put_Char(C : Character) is
   begin
      LCD_Hi(LCD_RS);
      LCD_Write_4x2Bits(C);
   end LCD_Put_Char;

   procedure LCD_Put_String(S : String) is
   begin
      for i in S'Range loop
         LCD_Put_Char(S(i));
      end loop;
   end LCD_Put_String;

   procedure LCD_Put_String(X : Byte; Y : Byte; S : String) is
   begin
      LCD_Goto(X, Y);
      LCD_Put_String(S);
   end LCD_Put_String;

end LCD;
