-- Timer package

with Ada.Interrupts.Names;
with Registers;                   use Registers;
with STM32F4;                     use STM32F4;
with STM32F4.Timer;

package body MUX_Timer is

   -- the 4 digit enable bits, in the same order as Digit_Vals below
   Digit_Bit : constant array(DIGIT) of Word := (16#1000#, 16#2000#, 16#4000#, 16#8000#);

   -- the 4 digits to be displayed.  These are indexes into array Segs below
   Digit_Vals  : array(DIGIT) of SEG_INDEX := (others => SEG_INDEX'Last);

   -- segment patterns for the 10 digits and Blank (Segment A is LSB)
   Segs  : constant array(SEG_INDEX) of Byte := (16#3F#, 16#06#, 16#5B#, 16#4F#, 16#66#,
						 16#6D#, 16#7D#, 16#07#, 16#7F#, 16#6F#,
					         16#00#);

   protected TimerP is
      pragma Interrupt_Priority;

   private
      -- which digit is now being displayed
      Cur_Digit : DIGIT := 0;

      procedure Interrupt_Handler;
      pragma Attach_Handler
         (Interrupt_Handler,
          Ada.Interrupts.Names.TIM6_DAC_Interrupt);

   end TimerP;

   protected body TimerP is

      -- handler for Timer 6
      procedure Interrupt_Handler is
	 -- these clear the segment outputs (spread over 3 ports!)
	 Clear_B   : constant Word := 16#03_0000#;
	 Clear_C   : constant Word := 16#06_0000#;
	 Clear_D   : constant Word := 16#CC_0000#;
	 B8        : Bits_8;
	 W32       : Bits_32;
      begin
         --  Clear interrupt
	 TIM6.SR := TIM6.SR - STM32F4.Timer.UIF;
	 -- turn off current digit
	 GPIOB.BSRR := Shift_Left(Digit_Bit(Cur_Digit), 16);
	 -- clear all segment outputs
	 GPIOB.BSRR := Clear_B;
	 GPIOC.BSRR := Clear_C;
	 GPIOD.BSRR := Clear_D;
	 -- advance to next digit (3 rolls over to 0)
	 Cur_Digit := Cur_Digit + 1;
	 -- get next segment pattern in form of boolean array
	 B8 := B8_From_B(Segs(Digit_Vals(Cur_Digit)));
	 -- set port C segment outputs
	 W32(0..31) := (others => False);
	 W32(1..2) := B8(6..7);
	 GPIOC.BSRR := W_From_B32(W32);
	 -- set port B segment outputs
	 W32(0..31) := (others => False);
	 W32(0..1) := B8(4..5);
	 GPIOB.BSRR := W_From_B32(W32);
	 -- set port D segment outputs
	 W32(0..31) := (others => False);
	 W32(6..7) := B8(2..3);
	 W32(2..3) := B8(0..1);
	 GPIOD.BSRR := W_From_B32(W32);
	 -- enable our new current digit
	 GPIOB.BSRR := Digit_Bit(Cur_Digit);
      end Interrupt_Handler;
   end TimerP;

   -- config Timer 6 to generate 2.5ms interrupts
   procedure Init is
   begin
      RCC.APB1ENR := RCC.APB1ENR or TIM6EN;
      TIM6.PSC  := CLK_US - 1;              -- 2x 42MHz
      TIM6.CR1  := STM32F4.Timer.URS;
--      TIM6.CR1  := STM32F4.Timer.ARPE or STM32F4.Timer.URS;
      TIM6.DIER := STM32F4.Timer.UIE;
      Set_Interrupt_Time(UPDATE_RATE'Last);
   end Init;

   -- enable Timer 6
   procedure Enable is
   begin
      TIM6.CR1  := TIM6.CR1 or STM32F4.Timer.CEN;  -- enable timer
   end Enable;

   procedure Set_Val(Value : DISP_VAL; Blank_LZ : Boolean) is
      Val       : Natural;
      Digit_Val : Natural;
      Digit     : Natural;
      BLZ       : Boolean := Blank_LZ;
   begin
      Val := Natural(Value);
     Digit_Val := Val / 1000;
      if (Digit_val = 0) and BLZ then
	 Digit := Blank;
      else
	 Digit := Digit_Val;
	 BLZ := False;
      end if;
      Digit_Vals(0) := SEG_INDEX(Digit);
      Val := Val mod 1000;

      Digit_Val := Val / 100;
      if (Digit_val = 0) and BLZ then
	 Digit := Blank;
      else
	 Digit := Digit_Val;
	 BLZ := False;
      end if;
      Digit_Vals(1) := SEG_INDEX(Digit);
      Val := Val mod 100;

      Digit_Val := Val / 10;
      if (Digit_val = 0) and BLZ then
	 Digit := Blank;
      else
	 Digit := Digit_Val;
      end if;
      Digit_Vals(2) := SEG_INDEX(Digit);
      Val := Val mod 10;

      Digit := Val;
      Digit_Vals(3) := SEG_INDEX(Digit);
   end Set_Val;

   procedure Set_Interrupt_Time(Per_Sec : UPDATE_RATE) is
   begin
      TIM6.ARR := HWord(((1_000_000/PRE_US)/(4 * Integer(Per_Sec))) - 1);
   end Set_Interrupt_Time;


end MUX_Timer;
